/**********************************************************************
 * $Id: httable.cpp,v 1.7 2002/04/15 18:13:34 julien Exp $
 *
 * Name:     httable.cpp
 * Project:  OWT HTML manipulation library
 * Purpose:  Implementation of the OWTHtTable class.
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1998-2000, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: httable.cpp,v $
 * Revision 1.7  2002/04/15 18:13:34  julien
 * Add Cell Attributes support and GetCol()/GetRow() fonctions.
 * Replace ',' bf ' ' in TextTable for Quatro (Bug 874)
 *
 * Revision 1.6  2000/11/29 20:30:34  daniel
 * Convert to ASCII instead of ISOLatin1 in OWTTextTable::Draw()
 *
 * Revision 1.5  2000/11/29 19:53:31  daniel
 * Convert HTML escaped chars to ISOLatin1 in OWTTextTable::Draw()
 *
 * Revision 1.4  1999/07/21 13:39:20  daniel
 * Implemented InsertRow()
 *
 * Revision 1.3  1999/07/14 04:25:24  daniel
 * Removed content-type
 *
 * Revision 1.2  1999/06/04 19:34:22  daniel
 * Improved SetCellRange() error message.
 *
 * Revision 1.1  1999/05/28 16:57:36  daniel
 * Initial revision
 *
 **********************************************************************/

#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "owtht.h"

/*=====================================================================
                        OWTHtTable class.
 =====================================================================*/

/**********************************************************************
 *                       OWTHtTable::OWTHtTable()
 *
 **********************************************************************/
OWTHtTable::OWTHtTable( )
            :OWTHtBlock("Table", "Table", NULL)
{
    m_numRows = m_numCols = 0;
    m_papapszCells = NULL;
    m_papapszCellAttributes = NULL;

    m_numWrapColumns = -1;  /* Do not wrap */
    m_nBorderWidth = 1;

    m_numHeaderRows = 1;  /* 1 row for column headers */
    m_numHeaderCols = 1;  /* 1 column for row headers */

    while(m_numRows < m_numHeaderRows)
        AddRow(NULL);
    while(m_numCols < m_numHeaderCols)
        AddColumn(NULL);
}

/**********************************************************************
 *                       OWTHtTable::~OWTHtTable()
 *
 **********************************************************************/
OWTHtTable::~OWTHtTable()
{
    if (m_numRows && m_papapszCells)
    {
        int i;
        for(i=0; i<m_numRows; i++)
            CSLDestroy(m_papapszCells[i]);

        CPLFree(m_papapszCells);
        CPLFree(m_papapszCellAttributes);
    }

}

/**********************************************************************
 *                       OWTHtTable::SetText()
 *
 * Set the text contents of the current object.  This will replace
 * any previous contents, and reset the object to a not-parsed state.
 *
 * This function is not currently supported.  When SetText() is
 * supported, then proper parsing of the HTML TABLE codes should
 * be implemented as well.
 **********************************************************************/
void OWTHtTable::SetText(char ** /*papszSrc*/, GBool /*bMakeCopy =TRUE */)
{
    CPLError(CE_Warning, CPLE_NotSupported,
             "OWTHtTable::SetText() has no effect.");
}

/**********************************************************************
 *                       OWTHtTable::Draw()
 *
 * Redraw the HTML table, handle wrapping after a given number of columns
 * if m_numWrapColumns has been set.
 **********************************************************************/
void OWTHtTable::Draw(FILE *fpOut /*=NULL*/)
{
    int iRow, iCol, nStartCol, nEndCol;

    /*-----------------------------------------------------------------
     * If reversed orientation requested, then pass control to DrawReversed()
     *----------------------------------------------------------------*/
    if (m_bReversedOrientation)
    {
        DrawReversed(fpOut);
        return;
    }

    if (fpOut == NULL)
        fpOut = stdout;

    nStartCol = m_numHeaderCols;  // Do not count the row header column

    while(nStartCol < m_numCols)
    {
        /*-------------------------------------------------------------
         * If wrapping is needed, then we produce one HTML sub-TABLE for
         * each set of columns
         *------------------------------------------------------------*/
        if (m_numWrapColumns > m_numHeaderCols)
        {
            nEndCol = nStartCol + m_numWrapColumns - m_numHeaderCols;
            if (nEndCol > m_numCols)
                nEndCol = m_numCols;
        }
        else
            nEndCol = m_numCols;

        fprintf(fpOut, "<TABLE BORDER=%d>\n", m_nBorderWidth);

        for(iRow=0; iRow<m_numRows; iRow++)
        {
            fprintf(fpOut, "<TR>\n");

            /*---------------------------------------------------------
             * Print the row header column(s)
             *--------------------------------------------------------*/
            for(iCol=0; iCol<m_numHeaderCols; iCol++)
            {
                if ( EQUAL(m_papapszCells[iRow][iCol], "") )
                    fprintf(fpOut, "  <TD>&nbsp;</TD>\n");
                else
                    fprintf(fpOut, "  <TD ALIGN=CENTER><B>%s</B></TD>\n", 
                                         m_papapszCells[iRow][iCol]);
            }

            /*---------------------------------------------------------
             * Print data cells, taking wrapping into account
             *--------------------------------------------------------*/
            for (iCol=nStartCol; iCol < nEndCol; iCol++)
            {
                    DrawTDWithAttributes(fpOut, iCol, iRow, 0);
            }

            fprintf(fpOut, "</TR>\n");
        }

        fprintf(fpOut, "</TABLE>\n");

        nStartCol = nEndCol;
    }
}

/**********************************************************************
 *                       OWTHtTable::DrawReversed()
 *
 * (Private Method)
 *
 * Redraw the HTML table, but with the rows as columns, and the
 * columns as rows.
 **********************************************************************/
void OWTHtTable::DrawReversed(FILE *fpOut /*=NULL*/)
{
    int iCol, iRow, nStartRow, nEndRow;

    if (fpOut == NULL)
        fpOut = stdout;

    nStartRow = m_numHeaderCols;  // Do not count the Col header Row

    while(nStartRow < m_numRows)
    {
        /*-------------------------------------------------------------
         * If wrapping is needed, then we produce one HTML sub-TABLE for
         * each set of Rows
         *------------------------------------------------------------*/
        if (m_numWrapColumns > m_numHeaderRows)
        {
            nEndRow = nStartRow + m_numWrapColumns - m_numHeaderRows;
            if (nEndRow > m_numRows)
                nEndRow = m_numRows;
        }
        else
            nEndRow = m_numRows;

        fprintf(fpOut, "<TABLE BORDER=%d>\n", m_nBorderWidth);

        for(iCol=0; iCol<m_numCols; iCol++)
        {
            fprintf(fpOut, "<TR>\n");

            /*---------------------------------------------------------
             * Print the Row header Columns
             *--------------------------------------------------------*/
            for(iRow=0; iRow<m_numHeaderRows; iRow++)
            {
                if ( EQUAL(m_papapszCells[iRow][iCol], "") )
                    fprintf(fpOut, "  <TD>&nbsp;</TD>\n");
                else
                    fprintf(fpOut, "  <TD ALIGN=CENTER><B>%s</B></TD>\n", 
                                         m_papapszCells[iRow][iCol]);
            }

            /*---------------------------------------------------------
             * Print data cells, taking wrapping into account
             *--------------------------------------------------------*/
            for (iRow=nStartRow; iRow < nEndRow; iRow++)
            {
                    DrawTDWithAttributes(fpOut, iCol, iRow, 1);
            }

            fprintf(fpOut, "</TR>\n");
        }

        fprintf(fpOut, "</TABLE>\n");

        nStartRow = nEndRow;
    }
}


/**********************************************************************
 *                       OWTHtTable::SetCellAttribute()
 *
 * Set a specific attribute of a cell
 * Attributes:
 * 0 = ALIGN (default = center)
 * 1 = WIDTH
 * 2 = HEIGHT
 * 3 = BGCOLOR
 **********************************************************************/
int OWTHtTable::SetCellAttribute(int nCol, int nRow, char *pszValue)
{

    if (nRow >= m_numRows || nCol >= m_numCols)
    {
        CPLError(CE_Failure, CPLE_IllegalArg,
                 "SetCellAttribute(%d, %d): Cell out of range, table is only %d x %d.",
                 nRow, nCol, m_numCols, m_numRows);
        return -1;
    }
    else
    {
        while(nRow >= m_numRows)
            if (AddRow(NULL) != 0) return -1;
        while(nCol >= m_numCols)
            if (AddColumn(NULL) != 0) return -1;
    }

    CPLFree(m_papapszCellAttributes[nRow][nCol]);
    m_papapszCellAttributes[nRow][nCol] = CPLStrdup(pszValue);

    return 0;
}


/**********************************************************************
 *                       OWTHtTable::DrawTDWithAttributes()
 *
 * print <TD ATTR1=... ATTR2=... ...>
 **********************************************************************/
void OWTHtTable::DrawTDWithAttributes(FILE *fpOut, int nCol, int nRow,
                                      bool bReversed)
{
    fprintf(fpOut, "  <TD ALIGN=CENTER");

    if(!EQUAL(m_papapszCellAttributes[nRow][nCol], ""))
        fprintf(fpOut, "%s", m_papapszCellAttributes[nRow][nCol]);

    fprintf(fpOut, ">");

    if ( EQUAL(m_papapszCells[nRow][nCol], "") )
        fprintf(fpOut, "  &nbsp;</TD>\n");
    else
        if(bReversed)
        {
            if (nCol < m_numHeaderCols)
                fprintf(fpOut, "<B>%s</B></TD>\n", 
                        m_papapszCells[nRow][nCol]);
            else
                fprintf(fpOut, "%s</TD>\n", 
                        m_papapszCells[nRow][nCol]);
        }
        else
        {
            if (nRow < m_numHeaderRows)
                fprintf(fpOut, "<B>%s</B></TD>\n", 
                        m_papapszCells[nRow][nCol]);
            else
                fprintf(fpOut, "%s</TD>\n", 
                        m_papapszCells[nRow][nCol]);
        }
}


/**********************************************************************
 *                       OWTHtTable::AddRow()
 *
 * Add a row to the table, and initialize the cell values if 
 * papszData!=NULL
 * Note that if the values in papszData are used, then they will be 
 * copied, and papszData will still be owned by the caller.
 *
 * Returns 0 on success, or -1 on error.
 **********************************************************************/
int OWTHtTable::AddRow(char **papszData)
{
    int     iCol, numSrcCols;

    numSrcCols = CSLCount(papszData);
    
    /*-----------------------------------------------------------------
     * Realloc Array of rows
     *----------------------------------------------------------------*/
    m_papapszCells = (char ***)CPLRealloc(m_papapszCells,
                                          (m_numRows+1)*sizeof(char**));
    m_numRows++;
    m_papapszCells[m_numRows-1] = (char **)CPLCalloc(m_numCols+1, 
                                                     sizeof(char*));

    m_papapszCellAttributes = (char ***)CPLRealloc(m_papapszCellAttributes,
                                          (m_numRows)*sizeof(char**));
    m_papapszCellAttributes[m_numRows-1] = (char **)CPLCalloc(m_numCols+1, 
                                                     sizeof(char*));

    /*-----------------------------------------------------------------
     * Init cell values for this new row
     *----------------------------------------------------------------*/
    for(iCol=0; iCol<m_numCols; iCol++)
    {
        if (iCol < numSrcCols)
            m_papapszCells[m_numRows-1][iCol] = CPLStrdup(papszData[iCol]);
        else
            m_papapszCells[m_numRows-1][iCol] = CPLStrdup("");

        m_papapszCellAttributes[m_numRows-1][iCol] = CPLStrdup("");
    }

    return 0;
}

/**********************************************************************
 *                       OWTHtTable::AddColumn()
 *
 * Add a column to the table, and initialize the cell values if
 * papszData!=NULL
 *
 * Returns 0 on success, or -1 on error.
 **********************************************************************/
int  OWTHtTable::AddColumn(char **papszData)
{
    int     iRow, numSrcRows;

    numSrcRows = CSLCount(papszData);
    
    /*-----------------------------------------------------------------
     * Add 1 cell to each row for the new column
     *----------------------------------------------------------------*/
    for(iRow=0; iRow<m_numRows; iRow++)
    {
        if (iRow < numSrcRows)
            m_papapszCells[iRow] = CSLAddString(m_papapszCells[iRow],
                                                papszData[iRow]);
        else
            m_papapszCells[iRow] = CSLAddString(m_papapszCells[iRow], "");

        m_papapszCellAttributes[iRow] = 
            CSLAddString(m_papapszCellAttributes[iRow], "");
    }

    m_numCols++;

    return 0;
}

/**********************************************************************
 *                       OWTHtTable::InsertRow()
 *
 * Insert a row to the table at the specified column position, and 
 * initialize the cell values if papszData!=NULL
 *
 * Note that if the values in papszData are used, then they will be 
 * copied, and papszData will still be owned by the caller.
 *
 * Returns 0 on success, or -1 on error.
 **********************************************************************/
int OWTHtTable::InsertRow(int nInsertAt, char **papszData)
{
    int     iRow, iCol, numSrcCols;

    numSrcCols = CSLCount(papszData);
    
    /*-----------------------------------------------------------------
     * Realloc Array of rows and shift all the rows that follow the new one
     *----------------------------------------------------------------*/
    m_papapszCells = (char ***)CPLRealloc(m_papapszCells,
                                          (m_numRows+1)*sizeof(char**));
    m_papapszCellAttributes = (char ***)CPLRealloc(m_papapszCellAttributes,
                                          (m_numRows+1)*sizeof(char**));
    m_numRows++;

    if (nInsertAt < 0)
        nInsertAt = 0;
    else if (nInsertAt >= m_numRows)
        nInsertAt = m_numRows -1;

    for(iRow = m_numRows-1; iRow > nInsertAt; iRow--)
    {
        m_papapszCells[iRow] = m_papapszCells[iRow-1];
        m_papapszCellAttributes[iRow] = m_papapszCellAttributes[iRow-1];
    }
    m_papapszCells[nInsertAt] = (char **)CPLCalloc(m_numCols+1, 
                                                   sizeof(char*));
    m_papapszCellAttributes[nInsertAt] = (char **)CPLCalloc(m_numCols+1, 
                                                   sizeof(char*));

    /*-----------------------------------------------------------------
     * Init cell values for this new row
     *----------------------------------------------------------------*/
    for(iCol=0; iCol<m_numCols; iCol++)
    {
        if (iCol < numSrcCols)
            m_papapszCells[nInsertAt][iCol] = CPLStrdup(papszData[iCol]);
        else
            m_papapszCells[nInsertAt][iCol] = CPLStrdup("");

        m_papapszCellAttributes[nInsertAt][iCol] = CPLStrdup("");
    }

    return 0;
}

/**********************************************************************
 *                       OWTHtTable::SetCell()
 *
 * Set the value for the specified cell.  
 * Create the new cell if necessary and bforceCreate=TRUE (default is FALSE)
 *
 * Keep in mind that row 0 and col 0 should be reserved for row and 
 * column headers.
 *
 * Returns 0 on success, or -1 on error.
 **********************************************************************/
int OWTHtTable::SetCell(int iCol, int iRow, const char *pszData, 
                       GBool bForceCreate /*= FALSE*/)
{

    if ( (iRow >= m_numRows || iCol >= m_numCols) && !bForceCreate)
    {
        CPLError(CE_Failure, CPLE_IllegalArg,
                 "SetCell(%d, %d): Cell out of range, table is only %d x %d.",
                 iRow, iCol, m_numCols, m_numRows);
        return -1;
    }
    else
    {
        while(iRow >= m_numRows)
            if (AddRow(NULL) != 0) return -1;
        while(iCol >= m_numCols)
            if (AddColumn(NULL) != 0) return -1;
    }

    CPLFree(m_papapszCells[iRow][iCol]);
    m_papapszCells[iRow][iCol] = CPLStrdup(pszData);

    return 0;
}

/**********************************************************************
 *                       OWTHtTable::SetCellRange()
 *
 * Set the value for all the cells in the specified range.  
 * Create the new cell if necessary and bforceCreate=TRUE (default is FALSE)
 *
 * When passing data for multiple rows, papszData should contain rows
 * of data to cover the whole range.
 *
 * Returns 0 on success, or -1 on error.
 **********************************************************************/
int  OWTHtTable::SetCellRange(int nCol1, int nRow1, int numCols, int numRows,
                              char **papszData, 
                              GBool bForceCreate /*= FALSE*/)
{
    int iRow, iCol, iData;

    if (CSLCount(papszData) != numCols*numRows)
    {
        CPLError(CE_Failure, CPLE_IllegalArg,
                 "SetCellRange(): Size of Data array (%d) is not equal to "
                 "size of destination range (%d x %d = %d).",
                 CSLCount(papszData), numCols, numRows, numCols*numRows);
        return -1;
    }

    iData=0;
    for(iRow=nRow1; iRow < nRow1+numRows; iRow++)
    {
        for(iCol=nCol1; iCol < nCol1+numCols; iCol++)
        {
            SetCell(iCol, iRow, papszData[iData++], bForceCreate);
        }
    }

    return 0;
}

/**********************************************************************
 *                       OWTHtTable::SetWrapMaxCols()
 *
 * Set the maximum number of columns from the table that can be
 * displayed at once.  Tables wider than the specified number of
 * columns will be drawn as several sub-tables.
 *
 * Pass a value of -1 to disable wrapping.
 **********************************************************************/
void  OWTHtTable::SetWrapMaxCols(int numCols)
{
    m_numWrapColumns = numCols;
}

/**********************************************************************
 *                       OWTHtTable::SetBorderWidth()
 *
 * Set the HTML TABLE BORDER value
 **********************************************************************/
void  OWTHtTable::SetBorderWidth(int nBorderWidth)
{
    m_nBorderWidth = nBorderWidth;
}

/**********************************************************************
 *                       OWTHtTable::SetReversedOrientation()
 *
 * Reverse table orientation: rows become column, and columns become rows
 *
 * Note: this will only have an effect at redraw time.  The internal
 * arrays (rows and columns) still keep the same orientation.
 **********************************************************************/
void  OWTHtTable::SetReversedOrientation(GBool bReverseOrientation)
{
    m_bReversedOrientation = bReverseOrientation;
}

/**********************************************************************
 *                       OWTHtTable::SetNumHeaderRows()
 *
 * Set the number of header rows (rows that will appear in bold, etc.)
 *
 * Note: if the table is displayed reversed, then the value passed here
 *       will define the number of header columns
 **********************************************************************/
void  OWTHtTable::SetNumHeaderRows(int numHeaderRows)
{
    m_numHeaderRows = numHeaderRows;
}

/**********************************************************************
 *                       OWTHtTable::SetNumHeaderCols()
 *
 * Set the number of header columns (columns that will appear in bold, etc.)
 *
 * Note: if the table is displayed reversed, then the value passed here
 *       will define the number of header rows
 **********************************************************************/
void  OWTHtTable::SetNumHeaderColumns(int numHeaderColumns)
{
    m_numHeaderCols = numHeaderColumns;
}

/**********************************************************************
 *                       OWTHtTable::GetRow()
 *
 * Return m_numRows (the number of row)
 **********************************************************************/
int  OWTHtTable::GetRow()
{
    return m_numRows;
}

/**********************************************************************
 *                       OWTHtTable::GetCol()
 *
 * Return m_numCols (the number of column)
 **********************************************************************/
int  OWTHtTable::GetCol()
{
    return m_numCols;
}



/*=====================================================================
                        OWTTextTable class.
 =====================================================================*/

/**********************************************************************
 *                       OWTTextTable::OWTTextTable()
 *
 **********************************************************************/
OWTTextTable::OWTTextTable( )
            :OWTHtTable()
{
    /*-----------------------------------------------------------------
     * "text/tab-separated-values" appears to be the official MIME type
     * for tab-delimited data files.  (.tsv extension)
     *----------------------------------------------------------------*/
    m_pszDelimiter   = CPLStrdup("\t");
}

/**********************************************************************
 *                       OWTTextTable::~OWTTextTable()
 *
 **********************************************************************/
OWTTextTable::~OWTTextTable()
{
    CPLFree(m_pszDelimiter);
}

/**********************************************************************
 *                       OWTTextTable::SetDelimiter()
 *
 * Set cell delimiter, default is tab.
 **********************************************************************/
void OWTTextTable::SetDelimiter(const char *pszDelimiter)
{
    CPLFree(m_pszDelimiter);
    m_pszDelimiter = CPLStrdup(pszDelimiter);
}

/**********************************************************************
 *                       OWTTextTable::Draw()
 *
 * Redraw the table.
 *
 * Note: We have to manually generate a CR/LF at the end of the lines 
 *       because on Unix, the Web Server does not include the CR, 
 *       and this confuses both Lotus and Quattro on Windows
 **********************************************************************/
void OWTTextTable::Draw(FILE *fpOut /*=NULL*/)
{
    int         iOutRow, iOutCol, nOutRows, nOutCols, nHdrCols, nHdrRows;
    const char szCRLF[3] = {13, 10, 0};
    char        *pszCell, *pszChr;

    if (fpOut == NULL)
        fpOut = stdout;

    /*-----------------------------------------------------------------
     * Reverse Table orientation if necessary
     *----------------------------------------------------------------*/
    if ( ! m_bReversedOrientation)
    {
        nOutRows = m_numRows;
        nOutCols = m_numCols;
        nHdrRows = m_numHeaderRows;
        nHdrCols = m_numHeaderCols;
    }
    else
    {
        nOutRows = m_numCols;
        nOutCols = m_numRows;
        nHdrRows = m_numHeaderCols;
        nHdrCols = m_numHeaderRows;
    }

    /*-----------------------------------------------------------------
     * Print data cells
     *----------------------------------------------------------------*/
    for(iOutRow=0; iOutRow<nOutRows; iOutRow++)
    {
        for (iOutCol=0; iOutCol < nOutCols; iOutCol++)
        {
            if ( ! m_bReversedOrientation)
                pszCell = m_papapszCells[iOutRow][iOutCol];
            else
                pszCell = m_papapszCells[iOutCol][iOutRow];

            /*---------------------------------------------------------
             * Replace any tabs inside the table values by spaces.
             *--------------------------------------------------------*/
            pszChr = pszCell;
            while( pszChr && m_pszDelimiter[0] != '\0' &&
                  (pszChr = strchr(pszChr, m_pszDelimiter[0])) != NULL)
                *pszChr = ' ';

            /*---------------------------------------------------------
             * Replace any ',' inside the table values by spaces. (for Quatro)
             *--------------------------------------------------------*/
            pszChr = pszCell;
            while( (pszChr = strchr(pszChr, ',')) != NULL )
                *pszChr = ' ';

            /*---------------------------------------------------------
             * Convert any HTML escaped chars (e.g. &eacute;) to ASCII
             *--------------------------------------------------------*/
            OWTDecodeHTML2ASCII(pszCell);

            /*---------------------------------------------------------
             * Row and Column header cell values are enclosed inside quotes.
             * YEP, there is going to be a problem if the value contains
             * one or more double quotes... we'll come back to fix that
             * one later.
             *--------------------------------------------------------*/
            if (iOutCol < nHdrCols || iOutRow < nHdrRows)
            {
                fprintf(fpOut, "%s\"%s\"", (iOutCol==0)?"":m_pszDelimiter,
                                           pszCell);
            }
            else
            {
                fprintf(fpOut, "%s%s", (iOutCol==0)?"":m_pszDelimiter, 
                                        pszCell );
            }
        }
        fprintf(fpOut, "%s", szCRLF);
    }

    fflush(fpOut);
}


