/**********************************************************************
 * $Id: owtcgien.cpp,v 1.6 2003/01/31 04:10:36 daniel Exp $
 *
 * Name:     owtcgien.cpp
 * Project:  OWT base library
 * Purpose:  Implementation of the OWTCGIEnv class used to access 
 *           HTML FORM parameters and CGI env. vars
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1998-2000, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: owtcgien.cpp,v $
 * Revision 1.6  2003/01/31 04:10:36  daniel
 * Added OWTCGIEnv::InitFromQueryString() and modified OWTParseForm()
 * to accept QUERY_STRING parameter to bypass env. var.
 *
 * Revision 1.5  2000/11/29 20:29:26  daniel
 * Added OWTDecodeHTML2ASCII()
 *
 * Revision 1.4  1999/05/28 16:58:16  daniel
 * Added SetEnv(), support for building URLs, and for accentuated chars
 *
 * Revision 1.3  1998/12/15 21:13:30  daniel
 * Fixed UnescapeString()
 *
 * Revision 1.2  1998/12/09 02:08:00  daniel
 * Converted to use portability library
 *
 * Revision 1.1  1998/12/04 19:15:17  daniel
 * Initial revision
 *
 *
 **********************************************************************/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "cpl_string.h"
#include "owtcgien.h"

/*=====================================================================
                         OWTCGIEnv class

OWTCGIEnv();
~OWTCGIEnv();

const char 	*GetFormParam(const char *pszName);
char      	**GetFormParamMulti(const char *pszName);
const char 	*GetEnv(const char *pszname);

 =====================================================================*/

/**********************************************************************
 *                       OWTCGIEnv::OWTCGIEnv()
 *
 **********************************************************************/
OWTCGIEnv::OWTCGIEnv()
{
    m_papszParams = NULL;
    m_bParsed = FALSE;
    m_papszEnvVars = NULL;
}

/**********************************************************************
 *                       OWTCGIEnv::~OWTCGIEnv()
 *
 **********************************************************************/
OWTCGIEnv::~OWTCGIEnv()
{
    CSLDestroy(m_papszParams);
    CSLDestroy(m_papszEnvVars);
}

/**********************************************************************
 *                       OWTCGIEnv::ParseForm()
 *
 * Use OWTParseForm() to parse the CGI form parameters from either the
 * GET or POST method and store the result inside class members ready
 * to be accessed by GetFormParam() and GetFormParamMulti().
 *
 * This method does not need to be called explicitly.  It is automatically
 * called by the class methods that need this information.
 **********************************************************************/
void OWTCGIEnv::ParseForm()
{
    if (m_bParsed == FALSE)
    {
        m_papszParams = OWTParseForm(NULL);
        m_bParsed = TRUE;
    }
}

/**********************************************************************
 *                       OWTCGIEnv::GetFormParam()
 *
 * Look for the form parameter with the specified name, and
 * return a static reference to its value or "" if there is no parameter
 * with this name.  This function is guaranteed to never return a NULL.
 *
 * The caller should not attempt to free the returned string.
 **********************************************************************/
const char *OWTCGIEnv::GetFormParam(const char *pszName)
{
    const char *pszValue;

    /* Make sure form parameters have been parsed */
    if (!m_bParsed)
        ParseForm();

    /* Fetch value in stringlist ... */
    pszValue = CSLFetchNameValue(m_papszParams, pszName);

    if (pszValue == NULL)
        pszValue = "";

    return pszValue;
}

/**********************************************************************
 *                       OWTCGIEnv::GetFormParamMulti()
 *
 * Look for the form parameter with the specified name, and
 * return a StringList with one string for each of its values.  This
 * function should be used for multiple selection lists.
 *
 * The returned StringList will have to be freed using 
 * CSLDestroy().
 **********************************************************************/
char **OWTCGIEnv::GetFormParamMulti(const char *pszName)
{
    char **papszValues;

    /* Make sure form parameters have been parsed */
    if (!m_bParsed)
        ParseForm();

    /* Fetch value in stringlist ... */
    papszValues = CSLFetchNameValueMultiple(m_papszParams, pszName);

    return papszValues;
}

/**********************************************************************
 *                       OWTCGIEnv::GetAllFormParams()
 *
 * Returns a stringlist with all the form parameter values.  Each line
 * contains a "NAME=VALUE" pair.
 *
 * Returns a reference to the stringlist that is used and maintained by
 * this object, and thus it should NOT be MODIFIED or destroyed by the caller.
 **********************************************************************/
char **OWTCGIEnv::GetAllFormParams()
{
    /* Make sure form parameters have been parsed */
    if (!m_bParsed)
        ParseForm();

    return m_papszParams;
}


/**********************************************************************
 *                       OWTCGIEnv::GetEnv()
 *
 * Safe cover function for the system getenv() that is guaranteed to 
 * never return NULL.  If a variable does not exist then "" is returned.
 *
 * This function should be used instead of getenv() if you want to use
 * the InitFromFile() method to restore the CGI environment when debugging.
 **********************************************************************/
const char *OWTCGIEnv::GetEnv(const char *pszName)
{
    const char *pszValue=NULL;

    /* First look in the local array of Env Vars values
     * loaded by InitFromFile()... if not found then the system
     * getenv() will be called.
     */
    if (m_papszEnvVars)
        pszValue = CSLFetchNameValue(m_papszEnvVars, pszName);

    if (pszValue == NULL)
    {
        pszValue = getenv(pszName);
        if (pszValue == NULL)
            pszValue = "";
    }

    return pszValue;
}

/**********************************************************************
 *                       OWTCGIEnv::SetEnv()
 *
 * Safe method to set env. vars for use with OWTCGIEnv::GetEnv().
 *
 * This function should be used instead of putenv() if you want to use
 * the InitFromFile() method to restore the CGI environment when debugging.
 **********************************************************************/
void OWTCGIEnv::SetEnv(const char *pszName, const char *pszValue)
{
    m_papszEnvVars = CSLSetNameValue(m_papszEnvVars, pszName, pszValue);
}

/**********************************************************************
 *                       OWTCGIEnv::DumpToFile()
 *
 * For debugging purposes only.
 *
 * Use the DumpToFile() method to capture the state of the CGI Env.
 * request and store it into a file.
 *
 * Then use InitFromFile() to reload this environment and debug the 
 * application.
 **********************************************************************/
#define CGIINFO_SEPARATOR  "--- end params/start env vars ---"
void OWTCGIEnv::DumpToFile(const char *pszFname)
{
    char **papszInfo;

    /* Create a stringlist with all the info to save and call CSLSave()
     * The form parameters will come first, then a separator line, then
     * the environment variables.
     */

    if (!m_bParsed)
        ParseForm();

    papszInfo = CSLDuplicate(m_papszParams);

    papszInfo = CSLAddString(papszInfo, CGIINFO_SEPARATOR);

    /* ... add env. vars...
     */
    papszInfo = CSLAddNameValue(papszInfo, "SERVER_SOFTWARE", 
                                        GetEnv("SERVER_SOFTWARE"));
    papszInfo = CSLAddNameValue(papszInfo, "SERVER_NAME",
                                        GetEnv("SERVER_NAME"));
    papszInfo = CSLAddNameValue(papszInfo, "GATEWAY_INTERFACE",
                                        GetEnv("GATEWAY_INTERFACE"));
    papszInfo = CSLAddNameValue(papszInfo, "SERVER_PROTOCOL",
                                        GetEnv("SERVER_PROTOCOL"));
    papszInfo = CSLAddNameValue(papszInfo, "SERVER_PORT",
                                        GetEnv("SERVER_PORT"));
    papszInfo = CSLAddNameValue(papszInfo, "REQUEST_METHOD",
                                        GetEnv("REQUEST_METHOD"));
    papszInfo = CSLAddNameValue(papszInfo, "PATH_INFO",
                                        GetEnv("PATH_INFO"));
    papszInfo = CSLAddNameValue(papszInfo, "PATH_TRANSLATED",
                                        GetEnv("PATH_TRANSLATED"));
    papszInfo = CSLAddNameValue(papszInfo, "SCRIPT_NAME",
                                        GetEnv("SCRIPT_NAME"));
    papszInfo = CSLAddNameValue(papszInfo, "QUERY_STRING",
                                        GetEnv("QUERY_STRING"));
    papszInfo = CSLAddNameValue(papszInfo, "REMOTE_HOST",
                                        GetEnv("REMOTE_HOST"));
    papszInfo = CSLAddNameValue(papszInfo, "REMOTE_ADDR",
                                        GetEnv("REMOTE_ADDR"));
    papszInfo = CSLAddNameValue(papszInfo, "AUTH_TYPE",
                                        GetEnv("AUTH_TYPE"));
    papszInfo = CSLAddNameValue(papszInfo, "REMOTE_USER",
                                        GetEnv("REMOTE_USER"));
    papszInfo = CSLAddNameValue(papszInfo, "REMOTE_IDENT",
                                        GetEnv("REMOTE_IDENT"));
    papszInfo = CSLAddNameValue(papszInfo, "CONTENT_TYPE",
                                        GetEnv("CONTENT_TYPE"));
    papszInfo = CSLAddNameValue(papszInfo, "CONTENT_LENGTH",
                                        GetEnv("CONTENT_LENGTH"));

    /* Save info to file */
    CSLSave(papszInfo, pszFname);

    CSLDestroy(papszInfo);
}

/**********************************************************************
 *                       OWTCGIEnv::InitFromFile()
 *
 **********************************************************************/
void OWTCGIEnv::InitFromFile(const char *pszFname)
{
    char    **papszInfo, **ppszLine;
    int     nCount, nSepLine;
    GBool   bFound = FALSE;

    papszInfo = CSLLoad(pszFname);
    nCount = CSLCount(papszInfo);

    if (nCount)
    {
        /* Flush any existing values */
        CSLDestroy(m_papszEnvVars);
        m_papszEnvVars = NULL;
        CSLDestroy(m_papszParams);
        m_papszParams = NULL;
        m_bParsed = TRUE;

        /* Locate the position of the separator line */
        ppszLine = papszInfo;
        for(nSepLine=0; *ppszLine!=NULL; ppszLine++, nSepLine++)
        {
            if (EQUAL(*ppszLine, CGIINFO_SEPARATOR))
            {
                /* Found it ... extract the Env. Vars section */
                papszInfo = CSLRemoveStrings(papszInfo, nSepLine+1,
                                             nCount-(nSepLine+1), 
                                             &m_papszEnvVars);

                /* Extract the Params section */
                papszInfo = CSLRemoveStrings(papszInfo, nSepLine, 1, NULL);
                m_papszParams = papszInfo;

                /* Terminate the loop. */
                bFound = TRUE;
                break;
            }
        }
    }

    /* If no separator line found, then assume that the file contained
     * only a list of form parameters.
     */
    if (!bFound)
        m_papszParams = papszInfo;

}

/**********************************************************************
 *                       OWTCGIEnv::InitFromQueryString()
 *
 **********************************************************************/
void OWTCGIEnv::InitFromQueryString(const char *pszQueryString)
{
    if (m_bParsed)
    {
        CSLDestroy(m_papszParams);
        m_papszParams = NULL;
    }

    SetEnv("QUERY_STRING", pszQueryString);

    m_papszParams = OWTParseForm(pszQueryString);
    m_bParsed = TRUE;
}

/*=====================================================================
 *                         C Functions
 *
 * OWTParseForm() has been implemented in C so that it can be eventually
 * used by itself in a C application if we ever need to.
 *====================================================================*/

static void _OWTUnescapeString(char *pszSrc);
static char *_OWTEscapeString(const char *pszSrc);

/**********************************************************************
 *                       OWTParseForm()
 *
 * Returns a stringlist of "name: value" pairs containing all the 
 * current form input parameters.  The CGI environment variables are
 * checked to establish the CONTENT_LENGTH, and the REQUEST_METHOD that
 * was used.
 *
 * Passing a non-NULL value for pszQueryString will override the environment
 * variables checking and will use this value as QUERY_STRING instead.
 *
 * Reutrns a StringList that should be freed using CSLDestroy()
 * or NULL if the form parameters could not be read.
 **********************************************************************/
char **OWTParseForm(const char *pszQueryString)
{
    const char *pszMethod, *pszLength;
    int  nLength;
    char *pszInBuf;
    char **papszParams, **ppszLine;

    /* Read input based on the REQUEST_METHOD and CONTENT_LENGTH
     */
    pszMethod = getenv("REQUEST_METHOD");
    pszLength = getenv("CONTENT_LENGTH");

    if (pszQueryString == NULL &&
        pszMethod && strcmp(pszMethod, "POST") == 0 &&
        pszLength && (nLength = atoi(pszLength)) > 0)
    {
        /* POST method: allocate a buffer and read from stdin
         */
        pszInBuf = (char *)calloc(nLength + 1, sizeof(char));
        nLength = fread(pszInBuf, sizeof(char), nLength, stdin);
    }
    else if (pszQueryString)
    {
        /* Using QUERY_STRING passed as argument (bypass getenv() call)
         */
        pszInBuf = CPLStrdup(pszQueryString);
    }
    else
    {
        /* Using GET method ... read QUERY_STRING 
         */
        pszInBuf = getenv("QUERY_STRING");
        if (pszInBuf == NULL)
            pszInBuf = "";
        pszInBuf = CPLStrdup(pszInBuf);
    }

    /* Parse input... papszParams will contain a stringlist with 
     * "name=value" entries.
     */
    papszParams = CSLTokenizeStringComplex(pszInBuf, "&", FALSE, TRUE);
    
    CPLFree(pszInBuf);
    
    /* Unescape each line... since a unescaped line cannot be longer
     * than the escaped version, we don't need to reallocate copies of each
     * line... we just work on the same buffers instead.
     */
    if (papszParams)
    {
        for(ppszLine = papszParams; *ppszLine!=NULL; ppszLine++)
        {
            _OWTUnescapeString(*ppszLine);
        }
    }

    return papszParams;
}

/**********************************************************************
 *                       OWTBuildURL()
 *
 * Build a proiperly escaped URL with the provided URL path, and
 * array of FORM parameters.  papszParams[] should be a stringlist
 * of NAME=VALUE pairs.
 *
 * Returns a newly allocated string that should be freed using CPLFree()
 * or NULL if a problem happened.
 **********************************************************************/
char *OWTBuildURL(const char *pszURLPath, char **papszParams)
{
    int   nTotLen, iParam;
    char  *pszNewURL, *pszEscStr, *pSeparator, *pszEqualSign;

    if (pszURLPath == NULL)
        pszURLPath = "";

    /*-----------------------------------------------------------------
     * Compute the biggest possible size of the resulting URL
     *----------------------------------------------------------------*/
    nTotLen = strlen(pszURLPath) + 1;

    for(iParam=0; papszParams && papszParams[iParam]; iParam++)
    {
        nTotLen += strlen(papszParams[iParam])*3 + 1;
    }

    /*-----------------------------------------------------------------
     * Build URL, escaping form parameters as we go
     *----------------------------------------------------------------*/
    pszNewURL = (char*)CPLCalloc(nTotLen + 1, sizeof(char));
    strcpy(pszNewURL, pszURLPath);
    pSeparator = "?";
    for(iParam=0; papszParams && papszParams[iParam]; iParam++)
    {
        strcat(pszNewURL, pSeparator);
        pSeparator = "&";
        pszEqualSign = strchr(papszParams[iParam], '=');
        if (pszEqualSign)
            *pszEqualSign = '\0'; /* Just temporary... */

        pszEscStr = _OWTEscapeString(papszParams[iParam]);
        strcat(pszNewURL, pszEscStr);
        strcat(pszNewURL, "=");
        CPLFree(pszEscStr);

        if ( pszEqualSign )
        {
            pszEscStr = _OWTEscapeString(pszEqualSign+1);
            strcat(pszNewURL, pszEscStr);
            CPLFree(pszEscStr);
            *pszEqualSign = '=';  /* Restore the '=' in the original */
        }
    }
    
    return pszNewURL;
}

/**********************************************************************
 *                       _OWTUnescapeString()
 *
 * Do the replacements of escaped characters in a  query string.
 * Since a unescaped line cannot be longer
 * than the escaped version, we don't need to reallocate copies of each
 * line... we just work on the same buffers instead.
 **********************************************************************/
static void _OWTUnescapeString(char *pszStr)
{
    int     iIn=0, iOut=0;
    char    c;

    while(pszStr[iIn] != '\0')
    {
        if (pszStr[iIn] == '+')
        {
            pszStr[iOut++] = ' ';
            iIn++;
        }
        else if (pszStr[iIn] == '%' && pszStr[iIn+1]!='\0' && pszStr[iIn+2]!='\0')
        {
            c = toupper(pszStr[iIn+1]);
            pszStr[iOut] = 16*(c- ((c>='0'&&c<='9')?'0':('A'-10)) );

            c = toupper(pszStr[iIn+2]);
            pszStr[iOut] += (c- ((c>='0'&&c<='9')?'0':('A'-10)) );

            iIn+=3;
            iOut++;
        }
        else
        {
            pszStr[iOut++] = pszStr[iIn++];
        }
    }
    pszStr[iOut++] = '\0';
}

/**********************************************************************
 *                       _OWTEscapeString()
 *
 * Do the replacements of escaped characters in a  query string.
 * Since a escaped string will usually be longer than the original version, 
 * we CPLMalloc() a new string to return (but the original remains untouched).
 *
 * This function should not be called with the full QUERY_STRING, since 
 * the '?', '&' and '=' separators would be escaped... and we don't want
 * that.  It should be called once for each component of the QUERY_STRING 
 * instead.
 *
 * The newly allocated string is returned, and the source string is not
 * freed, so both string become the responsibility of the caller.
 *
 * According to RFC2396, all alphanumeric characters, plus the following
 * ones:  "-_.!~*'()"  are unreserved and do not need to be escaped. 
 *
 * Note: This function is only suitable to escape QUERY_STRING components,
 * passing a URL through it would result in escaping the ':', '/', ...
 * and other reserved characters and the URL would become invalid.
 **********************************************************************/

static char *_OWTEscapeString(const char *pszStrIn)
{
    int     iIn=0, iOut=0;
    char    c, *pszStrOut;

    if (pszStrIn == NULL)
        return NULL;

    /*-----------------------------------------------------------------
     * Alloc a new string buffer with the biggest possible escaped size
     *----------------------------------------------------------------*/
    pszStrOut = (char *)CPLMalloc((strlen(pszStrIn)*3+1) * sizeof(char));

    while(pszStrIn[iIn] != '\0')
    {
        c = pszStrIn[iIn];

        if (isalnum(pszStrIn[iIn]) || 
            strchr("-_.!~*'()", pszStrIn[iIn]) != NULL)
        {
            pszStrOut[iOut++] = pszStrIn[iIn];
            iIn++;
        }
        else if (pszStrIn[iIn] == ' ')
        {
            pszStrOut[iOut++] = '+';
            iIn++;
        }
        else
        {
            sprintf(pszStrOut + iOut, "%%%02X", (unsigned char)pszStrIn[iIn]);
            iOut += 3;
            iIn++;
        }
    }

    pszStrOut[iOut++] = '\0';

    return pszStrOut;
}


/**********************************************************************
 *                       OWTDecodeHTML()
 *
 * Convert a string containing escaped characters as defined in the
 * HTML 3.2 standard (i.e. "&name;" form) to their equivalent based on
 * the szCodes argument.
 *
 * szCodes[] contains each char followed by its HTML equivalent.
 * We will use strstr() to locate the codes in the string.
 *
 * This function modifies the source string buffer directly.
 **********************************************************************/
#define MAX_CODE_LEN  20
void  OWTDecodeHTMLCharSet(char *pszInStr, const char *pszCodes)
{
    unsigned char szThisCode[MAX_CODE_LEN+1], *pszCodeFound, *pszStr;
    int  iSrc, iDst, iCode;
    GBool bFoundCode;

    pszStr = (unsigned char *)pszInStr;

    for(iSrc=iDst=0; pszStr && pszStr[iSrc]; iSrc++)
    {
        if (pszStr[iSrc] == '&')
        {
            /*--------------------------------------------------------
             * Found a coded char... extract its name
             *-------------------------------------------------------*/
            bFoundCode = FALSE;
            for(iCode=0; !bFoundCode && 
                          iCode<MAX_CODE_LEN  &&        
                          pszStr[iSrc+iCode] != '\0'; iCode++)
            {
                szThisCode[iCode] = pszStr[iSrc+iCode];
                if (szThisCode[iCode] == ';')
                    bFoundCode = TRUE;
            }
            szThisCode[iCode] = '\0';

            if (bFoundCode && szThisCode[1] == '#')
            {
                pszStr[iDst++] = (unsigned char)atoi((char*)szThisCode+1);
                iSrc += strlen((char*)szThisCode)-1;
            }
            else if (bFoundCode &&
                     (pszCodeFound = (unsigned char*)strstr(pszCodes, 
                                                   (char*)szThisCode)) != NULL)
            {
                pszStr[iDst++] = *(pszCodeFound-1);
                iSrc += strlen((char*)szThisCode)-1;
            }
            else
            {
                if (iSrc != iDst)
                    pszStr[iDst] = pszStr[iSrc];

                iDst++;
            }
        }
        else
        {
            /*--------------------------------------------------------
             * Regular char... not much to do!
             *-------------------------------------------------------*/
            if (iSrc != iDst)
                pszStr[iDst] = pszStr[iSrc];

            iDst++;
        }
    }

    if (iDst != iSrc)
        pszStr[iDst] = '\0';
}


/**********************************************************************
 *                       OWTDecodeHTML2ISOLatin1()
 *
 * Convert a string containing escaped characters as defined in the
 * HTML 3.2 standard (i.e. "&name;" form) to their ISO-8859-1 equivalent
 *
 * This function modifies the source string buffer directly.
 **********************************************************************/
void  OWTDecodeHTML2ISOLatin1(char *pszInStr)
{
    /*-----------------------------------------------------------------
     * szCodes[] contains each char followed by its HTML equivalent.
     * We will use strstr() to locate the codes in the string.
     *----------------------------------------------------------------*/
    const char szCodes[] = "\"&quot;"  "&&amp;"  "<&lt;"  ">&gt;"
                           " &nbsp;"   
                           "\351&eacute;"  "\311&Eacute;" 
                           "\350&egrave;"  "\310&Egrave;" 
                           "\352&ecirc;"   "\312&Ecirc;" 
                           "\353&euml;"    "\313&Euml;" 
                           "\340&agrave;"  "\300&Agrave;" 
                           "\342&acirc;"   "\302&Acirc;" 
                           "\364&ocirc;"   "\324&Ocirc;" 
                           "\371&ugrave;"  "\331&Ugrave;" 
                           "\373&ucirc;"   "\333&Ucirc;" 
                           "\356&icirc;"   "\316&Icirc;" 
                           "\357&iuml;"    "\317&Iuml;" 
                           "\347&ccedil;"  "\307&Ccedil;" ;

    OWTDecodeHTMLCharSet(pszInStr, szCodes);
}


/**********************************************************************
 *                       OWTDecodeHTML2ASCII()
 *
 * Convert a string containing escaped characters as defined in the
 * HTML 3.2 standard (i.e. "&name;" form) to their 7-bit ASCII equivalent
 *
 * This function modifies the source string buffer directly.
 **********************************************************************/
void  OWTDecodeHTML2ASCII(char *pszInStr)
{
    /*-----------------------------------------------------------------
     * szCodes[] contains each char followed by its HTML equivalent.
     * We will use strstr() to locate the codes in the string.
     *----------------------------------------------------------------*/
    const char szCodes[] = "\"&quot;"  "&&amp;"  "<&lt;"  ">&gt;"
                           " &nbsp;"   
                           "e&eacute;"  "E&Eacute;" 
                           "e&egrave;"  "E&Egrave;" 
                           "e&ecirc;"   "E&Ecirc;" 
                           "e&euml;"    "E&Euml;" 
                           "a&agrave;"  "A&Agrave;" 
                           "a&acirc;"   "A&Acirc;" 
                           "o&ocirc;"   "O&Ocirc;" 
                           "u&ugrave;"  "U&Ugrave;" 
                           "u&ucirc;"   "U&Ucirc;" 
                           "i&icirc;"   "I&Icirc;" 
                           "i&iuml;"    "I&Iuml;" 
                           "c&ccedil;"  "C&Ccedil;" ;

    OWTDecodeHTMLCharSet(pszInStr, szCodes);
}
