/**********************************************************************
 * $Id: owtchart.cpp,v 1.4 1999/05/28 21:25:05 daniel Exp $
 *
 * Name:     owtchart.cpp
 * Project:  OWT - Chart building block
 * Purpose:  Implementation of the OWTChart class.
 * Note:     This class uses the GifDraw (gd1.3) library for its rendering
 *           engine.
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1998, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: owtchart.cpp,v $
 * Revision 1.4  1999/05/28 21:25:05  daniel
 * Little changes to port to VC++
 *
 * Revision 1.3  1999/05/28 17:00:42  daniel
 * GetLegendSample() now takes cgi sample program path as argument
 *
 * Revision 1.2  1998/12/09 02:08:02  daniel
 * Converted to use portability library
 *
 * Revision 1.1  1998/12/04 19:15:18  daniel
 * Initial revision
 *
 * Revision 1.3  1998/11/06 03:50:10  daniel
 * First working version of the Charting Demo application
 *
 * Revision 1.2  1998/10/30 16:40:05  daniel
 * Added RCS header
 *
 **********************************************************************/

#include "cpl_string.h"
#include "owtchart.h"

#include "gdfontt.h"
#include "gdfonts.h"

#include <malloc.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#ifdef WIN32
#include <io.h>     /* _setmode()   */
#include <fcntl.h>  /* O_BINARY     */
#endif

/*=====================================================================
                            OWTChart class
 =====================================================================*/

/**********************************************************************
 *                       OWTChart::OWTChart()
 *
 **********************************************************************/

OWTChart::OWTChart()
{
    m_pValues = NULL;
    m_numLines = 0;
    m_numValPerLine = 0;

    m_nWidth = 400;
    m_nHeight = 300;

    m_numYAxisDiv = 5;
    m_bHorGrid = FALSE;
    m_bXAxisTickMarks = TRUE;
    m_bYAxisTickMarks = TRUE;

    m_dMinValue = 0.0;
    m_dMaxValue = 1.0;

    m_papszXValueLabels = NULL;
}

/**********************************************************************
 *                       OWTChart::~OWTChart()
 *
 ********************************************************************** */

OWTChart::~OWTChart()
{
    if (m_pValues != NULL)
        free(m_pValues);

    CSLDestroy(m_papszXValueLabels);
}

/**********************************************************************
 *                       OWTChart::SetSize()
 *
 * Set new chart size.
 ********************************************************************** */
void OWTChart::SetSize(int nWidth, int nHeight)
{
    if (nWidth > 20 && nHeight > 20)
    {
        m_nWidth = nWidth;
        m_nHeight = nHeight;
    }
}

/**********************************************************************
 *                       OWTChart::SetValues()
 *
 * Use this function to pass the array of values to the chart.
 *
 * SetValues() makes its own copy of the array of values.  For multiline
 * charts, the values for the first line come first in the array,
 * followed by the values for the second line, ...
 ********************************************************************** */
void OWTChart::SetValues(int numLines, int numValPerLine, double *pValues)
{
    int    i, nValues;
    double *pCurVal;

    m_numLines = numLines;
    m_numValPerLine = numValPerLine;

    nValues = numLines * numValPerLine;
    if (nValues == 0) 
        return;

    /* Create our own copy of the array of values */
    if (m_pValues != NULL)
        free(m_pValues);
    m_pValues = (double*)malloc(nValues*sizeof(double));
    
    memcpy(m_pValues, pValues, nValues*sizeof(double));

    /* Find the highest and lowest values */
    pCurVal = m_pValues;
    m_dMinValue = m_dMaxValue = *pCurVal;

    for (i=0; i<nValues; i++, pCurVal++)
    {
        if (*pCurVal < m_dMinValue)
            m_dMinValue = *pCurVal;
        else if (*pCurVal > m_dMaxValue)
            m_dMaxValue = *pCurVal;
    }

    if (m_dMinValue > 0.0 && m_dMaxValue > 0.0)
        m_dMinValue = 0.0;
    else if (m_dMaxValue < 0.0 && m_dMinValue < 0.0)
        m_dMaxValue = 0.0;

    /* ___ for now, do not allow ANY negative values...  since they are
     * considered as missing values.
     * This will have to be changed once we handle missing values 
     * properly.___
     */
    if (m_dMinValue < 0.0)
        m_dMinValue = 0.0;
}

/**********************************************************************
 *                       OWTChart::SetValues()
 *
 * Same as above, but takes a NULL-terminated array of strings as input.
 ********************************************************************** */
void OWTChart::SetValues(int numLines, int numValPerLine, 
                        char **papszValues)
{
    int    i, nValues;
    double *pValues;

    nValues = numLines * numValPerLine;
    if (nValues == 0 || papszValues == NULL) 
        return;

    /* Create an array of doubles and pass it to SetValues() above and
     * SetValues() will make its own copy of the array... I realize this
     * is not very efficient... but it will be simpler this way for now.
     */
    pValues = (double*)calloc(nValues,sizeof(double));
    for(i=0; i<nValues && papszValues[i] != NULL; i++)
        pValues[i] = atof(papszValues[i]);

    SetValues(numLines, numValPerLine, pValues);

    free(pValues);
}


/**********************************************************************
 *                       OWTChart::Redraw()
 *
 * (Private method)
 *
 * Draw the chart to the gdImage.
 ********************************************************************** */
void OWTChart::Redraw(gdImagePtr pImage)
{
    int         nColorWhite, nColorBlack, nColorRed;

    /* Select font to use based on image height */
    if (m_nHeight < 100)
        m_pFont = gdFontTiny;
    else
        m_pFont = gdFontSmall;

    /* Horizontal offsets */
    m_nXLeft = (int)((log10(m_dMaxValue)+1)+2)*m_pFont->w;
    m_nXRight = m_nWidth-5;

    /* Vertical Offsets (note that (0,0) is the top-left corner) */
    m_nYTop = (m_pFont->h * 2);
    m_nYBottom = m_nHeight - (m_pFont->h * 2);

    /* Allocate bg color (white) and a black fg color */
    nColorWhite = gdImageColorAllocate(pImage, 255, 255, 255);
    nColorBlack = gdImageColorAllocate(pImage, 0, 0, 0);
    nColorRed   = gdImageColorAllocate(pImage, 255, 0, 0);

    /* Draw the axis */
    gdImageLine(pImage, m_nXLeft, m_nYBottom, m_nXLeft, m_nYTop, nColorBlack);
    gdImageLine(pImage, m_nXLeft, m_nYBottom, m_nXRight, m_nYBottom, nColorBlack);

    /* Chart the values */
    ChartValues(pImage, nColorRed);

    /* Value labels, tick marks, grid, ... */
    ChartAnnotations(pImage, nColorBlack);
}


/**********************************************************************
 *                       OWTChart::ChartValues()
 *
 * (Private method)
 *
 * Display the chart values on the gdImage.  This function assumes that
 * some initialisation has previously been done by Redraw() 
 ********************************************************************** */
void OWTChart::ChartValues(gdImagePtr pImage, int /*nColor*/)
{
    double dDy, dYScale, dXSpacing, *pCurVal;
    int    iLine, iValue;
    int    nLineColor;

    if (m_pValues == NULL || m_numValPerLine < 2)
        return;

    /* Compute scaling factor to use to map chart values to pixels */
    dDy = m_dMaxValue - m_dMinValue;

    if (dDy != 0.0)
        dYScale = (m_nYBottom - m_nYTop)/(dDy);
    else
        dYScale = 1.0;

    dXSpacing = (double)(m_nXRight - m_nXLeft)/(m_numValPerLine-1.0);

    /* Loop to draw the values */
    pCurVal = m_pValues;
    for (iLine=0; iLine < m_numLines; iLine++)
    {
        /* Allocate a color for this line
         * This method should allow us to generate 16 different colors.
         */
        nLineColor = gdImageColorAllocate(pImage, 200*((iLine+1)&1),
                                                  200*((iLine+1)&2)/2,
                                                  200*((iLine+1)&4)/4);

        for (iValue=0; iValue < m_numValPerLine-1; iValue++, pCurVal++)
        {
            /* Skip negative values.  Eventually, we should add a better
             * way to handle "MISSING" values.
             */
            if (*pCurVal >=0 && *(pCurVal+1) >= 0)
              gdImageLine(pImage, (int)(m_nXLeft+(iValue*dXSpacing)), 
                                (int)(m_nYBottom-(*pCurVal-m_dMinValue)*dYScale),
                                (int)(m_nXLeft+((iValue+1)*dXSpacing)), 
                                (int)(m_nYBottom-(*(pCurVal+1)-m_dMinValue)*dYScale) ,
                                nLineColor);
        }

        pCurVal++;
    }
}

/**********************************************************************
 *                       OWTChart::ChartAnnotations()
 *
 * (Private method)
 *
 * Display the chart labels, grid lines, tick marks, etc. on the gdImage.
 * This function assumes that some initialisation has previously been 
 * done by Redraw() 
 ********************************************************************** */
void OWTChart::ChartAnnotations(gdImagePtr pImage, int nColor)
{
    int     i, nX, nY, nTickMarkSize;
    double  dXSpacing, dYSpacing, dDy;
    char    szBuffer[100];

    nTickMarkSize = m_pFont->w/2;

    /* Y axis labels, tick marks and horizontal grid */
    dDy = (m_dMaxValue - m_dMinValue)/(m_numYAxisDiv-1);
    dYSpacing = (double)(m_nYBottom - m_nYTop)/(double)(m_numYAxisDiv-1);

    for (i=0; i<m_numYAxisDiv; i++)
    {
        nY = (int)(m_nYBottom - i * dYSpacing);

        if (m_bYAxisTickMarks)
        {
            gdImageLine(pImage, m_nXLeft-nTickMarkSize, nY,
                                m_nXLeft, nY, nColor);
        }

        if (m_bHorGrid && i>0)
        {
            gdImageLine(pImage, m_nXLeft, nY,
                                m_nXRight, nY, nColor);
        }

        /* For now, always print integer value labels... 
         * this will have to be changed so that we prebuild the labels, and
         * take the size of the longest label to define the left offset.
         */
/*        sprintf(szBuffer, "%g", m_dMinValue + i*dDy);*/
        sprintf(szBuffer, "%d", (int)(m_dMinValue + i*dDy));

        gdImageString(pImage, m_pFont, 
                      m_nXLeft-(strlen(szBuffer))*m_pFont->w - nTickMarkSize*2, 
                      nY-m_pFont->h/2, (unsigned char*)szBuffer, nColor);
    }

    /* X Axis labels and tick marks */
    dXSpacing = (double)(m_nXRight-m_nXLeft)/(double)(m_numValPerLine-1);

    for (i=0; i<m_numValPerLine; i++)
    {
        nX = (int)(m_nXLeft + i * dXSpacing);

        if (m_bXAxisTickMarks)
        {
            gdImageLine(pImage, nX, m_nYBottom + nTickMarkSize,
                                nX, m_nYBottom, nColor);
        }

        /* If user provided X Axis value labels then use them, otherwise
         * generate some using the value indices (i.e. 1,2,3,4,5,6.... )
         *
         * To avoid label overlapping (with 50 years of data!) print only every
         * second label.
         */
        if (m_papszXValueLabels)
        {
            if (i%2 == 0 && m_papszXValueLabels[i])
            {
                strncpy(szBuffer, m_papszXValueLabels[i], 20);
                szBuffer[20] = '\0';
            }
            else
                szBuffer[0] = '\0';
        }
        else
            sprintf(szBuffer, "%d", i+1);

        gdImageString(pImage, m_pFont, 
                      nX - strlen(szBuffer)*m_pFont->w/2, 
                      m_nYBottom + nTickMarkSize*2,
                      (unsigned char *)szBuffer, nColor);
    }

}

/**********************************************************************
 *                       OWTChart::SaveSlideGIF()
 *
 ********************************************************************** */
int OWTChart::SaveSlideGIF(FILE *fpOut)
{
    gdImagePtr pImage;

    /* Allocate image in memory */
    pImage = gdImageCreate(m_nWidth, m_nHeight);

    /* Generate the chart */
    Redraw(pImage);

    /* Output the image (On WIN32 we have to make sure output file is 
     * in binary mode.)
     */
#ifdef WIN32
    if ( _setmode( _fileno( fpOut ), O_BINARY ) == -1)
        CPLError(CE_Failure, CPLE_FileIO, "OWTChart::SaveSlideGIF(): "
                      "Cannot set output file mode (%s)",  strerror(errno));
#endif
    gdImageGif(pImage, fpOut);
    fflush(fpOut);

    /* Destroy image in memory */
    gdImageDestroy(pImage);

    return 0;
}

/**********************************************************************
 *                       OWTChart::SaveSlideGIF()
 *
 * Same as above, but takes the name of a file to create as argument.
 ********************************************************************** */
int OWTChart::SaveSlideGIF(const char *pszFname)
{
    FILE *fpOut;
    int nStatus = 0;

    if ((fpOut = fopen(pszFname, "wb")) != NULL)
    {
        nStatus = SaveSlideGIF(fpOut);
        fclose(fpOut);
    }
    else
    {
        CPLError(CE_Failure, CPLE_FileIO, "OWTChart::SaveSlideGIF(%s): %s", 
                    pszFname, strerror(errno));
        nStatus = 1;
    }

    return nStatus;
}

/**********************************************************************
 *                       OWTChart::SetXValueLabels()
 *
 * Set strings to use for labels along the X axis.
 ********************************************************************** */
void OWTChart::SetXValueLabels(char **papszValueLabels)
{
    CSLDestroy(m_papszXValueLabels);
    m_papszXValueLabels = CSLDuplicate(papszValueLabels);
}


/**********************************************************************
 *                       OWTChart::GetLegendSamples()
 *
 * Return a MString with the URL for the legend sample for each line.
 ********************************************************************** */
char **OWTChart::GetLegendSamples(const char *pszGIFSampleURL,
                                  int nWidth, int nHeight)
{
    int     iLine;
    char    **papszSamples = NULL;

    for(iLine=0; iLine<m_numLines; iLine++)
    {
        papszSamples = CSLAppendPrintf(papszSamples, 
                                "%s/%d/%d/%d/%d/%d",
                                pszGIFSampleURL,
                                nWidth, nHeight, 200*((iLine+1)&1),
                                                 200*((iLine+1)&2)/2,
                                                 200*((iLine+1)&4)/4);
    }
    return papszSamples;
}
