/**********************************************************************
 * $Id: owtgdcharturl.cpp,v 1.4 2002/04/19 17:54:12 julien Exp $
 *
 * Name:     owtgdcharturl.cpp
 * Project:  OWT - Chart building block
 * Purpose:  Implementation of the OWTGDChartURL class.
 *           This class produces URLs for the OWT CGI chart engine.
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1998, 1999, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: owtgdcharturl.cpp,v $
 * Revision 1.4  2002/04/19 17:54:12  julien
 * Added GetParams function
 *
 * Revision 1.3  1999/07/27 20:10:42  daniel
 * Added possibility to have groups of lines the same color (for LCDC apps)
 *
 * Revision 1.2  1999/07/21 13:41:02  daniel
 * Modif. to properly handle pie chart legend samples in GetLegendSamples()
 *
 * Revision 1.1  1999/05/28 17:04:21  daniel
 * Initial revision
 *
 **********************************************************************/

#include <malloc.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "cpl_conv.h"
#include "cpl_string.h"

#include "owtgdchart.h"
#include "owtcgien.h"
#include "gdfontt.h"
#include "gdfonts.h"

/*=====================================================================
                            OWTGDChartURL class
 =====================================================================*/

/**********************************************************************
 *                       OWTGDChartURL::OWTGDChartURL()
 *
 **********************************************************************/

OWTGDChartURL::OWTGDChartURL(GDC_CHART_T nChartType)
{
    m_numLines = 0;
    m_numValPerLine = 0;

    m_papszParams = NULL;
    m_pszURLBuf = NULL;
    m_papszStyles = NULL;
    m_numLinesByColor = 1;

    SetChartType(nChartType);

    SetURLParam("W", "400");
    SetURLParam("H", "300");
}

/**********************************************************************
 *                       OWTGDChartURL::~OWTGDChartURL()
 *
 ********************************************************************** */

OWTGDChartURL::~OWTGDChartURL()
{
    CSLDestroy(m_papszParams);
    CPLFree(m_pszURLBuf);
    CSLDestroy(m_papszStyles);
}

/**********************************************************************
 *                       OWTGDChartURL::SetChartType()
 *
 * Set new chart type.  Use this function with care, since changing the
 * chart type after some other parameters values have been set may lead to 
 * unpredictable results.
 ********************************************************************** */
void OWTGDChartURL::SetChartType(GDC_CHART_T nChartType)
{
    m_eType = nChartType;

    SetURLParam("TYPE", (nChartType == GDC_LINE ? "LINE" :
                         nChartType == GDC_BAR ?  "BAR" :
                         nChartType == GDC_2DPIE ?  "PIE" :
                         nChartType == GDC_3DLINE ? "3DLINE" :
                         nChartType == GDC_3DBAR ?  "3DBAR" :
                         nChartType == GDC_3DPIE ?  "3DPIE" : 
                         nChartType == GDC_HBAR ?   "HBAR" :
                         nChartType == GDC_H2BAR ?  "H2BAR" :
                         nChartType == GDC_3DHBAR ? "3DHBAR" :
                         nChartType == GDC_3DH2BAR ?"3DH2BAR" :
                         nChartType == GDC_HILOCLOSE ?"HLC" :
                         nChartType == GDC_3DHILOCLOSE ?"3DHLC" : "???") );

}

/**********************************************************************
 *                       OWTGDChartURL::SetSize()
 *
 * Set new chart size.
 ********************************************************************** */
void OWTGDChartURL::SetSize(int nWidth, int nHeight)
{
    if (nWidth > 20 && nHeight > 20)
    {
        SetURLParam("W", nWidth);
        SetURLParam("H", nHeight);
    }
}

/**********************************************************************
 *                       OWTGDChartURL::SetValues()
 *
 * Use this function to pass the array of values to the chart.
 *
 * For multiline charts, the values for the first line come first in
 * the array, followed by the values for the second line, ...
 ********************************************************************** */
void OWTGDChartURL::SetValues(int numLines, int numValPerLine, double *pValues)
{
    int    i, nValues;
    char   **papszVals = NULL;

    nValues = numLines * numValPerLine;
    if (nValues == 0 || pValues == NULL) 
        return;

    /* Create a copy of the array of values to pass to SetURLParam() */
    for(i=0; i<nValues; i++)
        papszVals = CSLAppendPrintf(papszVals, "%g", pValues[i]);

    SetValues(numLines, numValPerLine, papszVals);

    CSLDestroy(papszVals);
}

/**********************************************************************
 *                       OWTGDChartURL::SetValues()
 *
 * Same as above, but takes a NULL-terminated array of strings as input.
 ********************************************************************** */
void OWTGDChartURL::SetValues(int numLines, int numValPerLine,
                              char **papszValues)
{
    int numLinesPerSet = (m_eType==GDC_H2BAR || m_eType==GDC_3DH2BAR)? 2:1;

    if (numLines == 0 || numValPerLine == 0 || papszValues == NULL) 
        return;

    m_numLines = numLines;
    m_numValPerLine = numValPerLine;

    SetURLParam("VALS", papszValues, '!');
    SetURLParam("NUMPTS", numValPerLine);
    SetURLParam("NUMSETS", numLines/numLinesPerSet);

}


/**********************************************************************
 *                       OWTGDChartURL::Redraw()
 *
 * (Private method)
 *
 * Generate the LINE or BAR chart
 ********************************************************************** */
void OWTGDChartURL::Redraw()
{
    int numClrSets = (m_eType == GDC_H2BAR || m_eType == GDC_3DH2BAR)? 2:1;

    /*----------------------------------------------------------------- 
     * Set colors for each line or set of bars 
     *----------------------------------------------------------------*/
    if ( m_numLines/numClrSets > 1)
    {
        char **papszColors = GetColorsArray(m_numLines/numClrSets,
                                            m_numLinesByColor);
        SetURLParam("SetColors", papszColors);
        CSLDestroy(papszColors);
    }

    /* Nothing else to do for now for LINE and BAR charts */

}

/**********************************************************************
 *                       OWTGDChartURL::RedrawPie()
 *
 * (Private method)
 *
 * Generate the PIE chart
 ********************************************************************** */
void OWTGDChartURL::RedrawPie()
{
    char        **papszColors = NULL;

    /*----------------------------------------------------------------- 
     * Set colors for each slice 
     *----------------------------------------------------------------*/
    papszColors = GetColorsArray(m_numValPerLine, m_numLinesByColor);
    SetURLParam("ExtColors", papszColors);
    CSLDestroy(papszColors);

    /*----------------------------------------------------------------- 
     * Set other style parameters
     *----------------------------------------------------------------*/
    //SetURLParam("EdgeColor", "000000" );
    //SetURLParam("3DAngle",   "45");
    //SetURLParam("3DDepth",   "15");
    //SetURLParam("PiePctLabels", "RIGHT");
    //SetURLParam("PieLabelLine", "1");
    //SetURLParam("PieLabelDist", "15");
}

/**********************************************************************
 *                       OWTGDChartURL::SaveSlideURL()
 *
 * Generates the URL for the chart.
 *
 * Returns a reference to an internal buffer that should not be freed
 * by the caller, and will remain valid until the next call.
 ********************************************************************** */
const char *OWTGDChartURL::SaveSlideURL(const char *pszCGIPath)
{

    //SetURLParam("BGColor",   "FFFFFF");  /* BG color   = white */
    //SetURLParam("PlotColor", "FFFFFF");  /* Bar Colors = blue  */

    if (m_eType == GDC_LINE  || m_eType == GDC_3DLINE ||
        m_eType == GDC_AREA  || m_eType == GDC_3DAREA ||
        m_eType == GDC_BAR   || m_eType == GDC_3DBAR  ||
        m_eType == GDC_HBAR  || m_eType == GDC_3DHBAR ||
        m_eType == GDC_H2BAR || m_eType == GDC_3DH2BAR||
        m_eType == GDC_HILOCLOSE || m_eType == GDC_3DHILOCLOSE) 
   {
        Redraw();
    }
    else if (m_eType == GDC_2DPIE || m_eType == GDC_3DPIE)
    {
        RedrawPie();
    }
    else
    {
        CPLError(CE_Failure, CPLE_IllegalArg, "Unsupported Chart type!");
    }
    
    /*----------------------------------------------------------------- 
     * Build the URL (and properly escape special chars)
     *----------------------------------------------------------------*/
    CPLFree(m_pszURLBuf);
    m_pszURLBuf = OWTBuildURL(pszCGIPath, m_papszParams);

    return m_pszURLBuf;;
}


/**********************************************************************
 *                       OWTGDChartURL::SetXValueLabels()
 *
 * Set strings to use for labels along the X axis.
 ********************************************************************** */
void OWTGDChartURL::SetXValueLabels(char **papszValueLabels)
{
    SetURLParam("XLabels", papszValueLabels);
}

/**********************************************************************
 *                       OWTGDChartURL::SetTitle()
 *
 * Set the main chart title
 ********************************************************************** */
void OWTGDChartURL::SetTitle(const char *pszTitle)
{
    SetURLParam("Title", pszTitle);
}

/**********************************************************************
 *                       OWTGDChartURL::SetXAxisTitle()
 *
 * Set the X axis title
 ********************************************************************** */
void OWTGDChartURL::SetXAxisTitle(const char *pszTitle)
{
    SetURLParam("XTitle", pszTitle);
}

/**********************************************************************
 *                       OWTGDChartURL::SetYAxisTitle()
 *
 * Set the Y axis title
 ********************************************************************** */
void OWTGDChartURL::SetYAxisTitle(const char *pszTitle)
{
    SetURLParam("YTitle", pszTitle);
}

/**********************************************************************
 *                       OWTGDChartURL::SetLeftRightTitle()
 *
 * Set the Left and Right sub-titles for H2BAR types
 ********************************************************************** */
void OWTGDChartURL::SetLeftRightTitles(const char *pszLTitle,
                                      const char *pszRTitle)
{
    /* There is a bug with I.E. 4.0 and Navigator 3.0, related to the
     * use of "&LTitle=value" inside a IMG SRC URL... the "&LT" is seen
     * as a "<" character by these browsers and this messes things up.
     * OWTChart still supports the use of "LTitle", but the use of 
     * "L_Title" is preferred.
     */
    SetURLParam("L_Title", pszLTitle);
    SetURLParam("R_Title", pszRTitle);
}

/**********************************************************************
 *                       OWTGDChartURL::SetLineStyles()
 *
 * Set line styles to use... makes sense only for line charts.
 * Default is all solid lines.
 *
 * papszStyles[] contains num_sets strings each defining one line
 * style in which spaces are transparent pixels, and any other character
 * turns on the corresponding pixel.
 *
 * If numLinesOfEachColor is set, then this will define the number of lines
 * that will share the same color... instead of having one color for each
 * line, there will be groups of lines with the same color, but quite
 * likely a different line style.
 *
 * Pass NULL to come back to the default (i.e. Solid lines).
 ********************************************************************** */
void OWTGDChartURL::SetLineStyles(char **papszStyles,
                                  int  numLinesOfEachColor /* = 1 */)
{
    /* We need to keep a copy for when we generate legend samples */
    CSLDestroy(m_papszStyles);
    m_papszStyles = CSLDuplicate(papszStyles);

    SetURLParam("LineStyles", papszStyles);

    m_numLinesByColor = MAX(1, numLinesOfEachColor);
}

/**********************************************************************
 *                       OWTGDChartURL::SetScaleType()
 *
 * Set the Y axis scale type: GDC_SCALE_LINEAR (the default) or GDC_SCALE_LOG
 ********************************************************************** */
void  OWTGDChartURL::SetScaleType(GDC_SCALE_T eScaleType)
{
    switch(eScaleType)
    {
      case GDC_SCALE_LINEAR:
        SetURLParam("YScaleType", "LIN");
        break;
      case GDC_SCALE_LOG:
        SetURLParam("YScaleType", "LOG");
        break;
    }
}



/**********************************************************************
 *                       OWTGDChartURL::GetLegendSamples()
 *
 * Return a MString with the URL for the legend sample for each line.
 ********************************************************************** */
char **OWTGDChartURL::GetLegendSamples(const char *pszGIFSampleURL,
                                       int nWidth, int nHeight)
{
    int     iLine, numStyles=0, numSamples;
    char    **papszColors=NULL, **papszSamples = NULL, **papszQueryStr=NULL;
    char    *pszPathInfo=NULL;

    if (pszGIFSampleURL == NULL)
        pszGIFSampleURL = "/cgi-bin/gifsample";

    /*----------------------------------------------------------------- 
     * Establish number of legend samples based on chart type
     *----------------------------------------------------------------*/
    if (m_eType == GDC_2DPIE || m_eType == GDC_3DPIE)
        numSamples = m_numValPerLine;
    else
        numSamples = m_numLines;

    numStyles = CSLCount(m_papszStyles);
    if (m_eType == GDC_LINE || numStyles > 0)
        papszQueryStr = CSLSetNameValue(NULL, "LineWidth", "2");

    papszColors = GetColorsArray(numSamples, m_numLinesByColor);

    papszSamples = (char**)CPLCalloc(numSamples+1, sizeof(char*));

    pszPathInfo = CPLStrdup(CPLSPrintf("%s/%d/%d",
                                       pszGIFSampleURL, nWidth, nHeight ));

    for(iLine=0; iLine<numSamples; iLine++)
    {
        if (m_eType == GDC_LINE && iLine < numStyles)
        {
            papszQueryStr = CSLSetNameValue(papszQueryStr, "LineStyle",
                                            m_papszStyles[iLine]);
        }

        papszQueryStr = CSLSetNameValue(papszQueryStr, "Color",
                                        papszColors[iLine]);

        papszSamples[iLine] = OWTBuildURL(pszPathInfo, papszQueryStr);
    }

    CPLFree(pszPathInfo);
    CSLDestroy(papszColors);
    CSLDestroy(papszQueryStr);

    return papszSamples;
}

/**********************************************************************
 *                       OWTGDChartURL::GetColorsArray()
 *
 * (Private Method)
 *
 * Return a MString with the Hex RGB values for the specified number 
 * of colors
 ********************************************************************** */
char **OWTGDChartURL::GetColorsArray(int numColors,
                                     int numLinesByColor /* = 1 */)
{
    static  char *apszDefColors[] = {"FF0000", "00FF00", "0000FF", "FF00FF",
                                    "00FFFF", "9966FF", "008800", "FF9944",
                                    "0088FF", "00FFCC", "CC3399", "66FFCC",
                                    "99FF33", "CCFF00", "FF3300", "99FF99",
                                    "66CCFF", "0099CC", "990033", "99FFFF",
                                    NULL };
    int     i, numDefColors;
    char    **papszColors = NULL;

    numLinesByColor = MAX(1, numLinesByColor);
    numDefColors = CSLCount(apszDefColors);

    for(i=0; i<numColors; i++)
    {
        papszColors = CSLAddString(papszColors,
                             apszDefColors[(i/numLinesByColor)%numDefColors]);
    }

    return papszColors;
}

/**********************************************************************
 *                       OWTGDChartURL::SetURLParam()
 *
 * Set a value in the list of URL parameters for hte OWTChart CGI engine.
 ********************************************************************** */
void OWTGDChartURL::SetURLParam(const char *pszName, const char *pszValue)
{
    m_papszParams = CSLSetNameValue(m_papszParams, pszName, pszValue);
}

void OWTGDChartURL::SetURLParam(const char *pszName, int nValue)
{
    char szBuf[50];

    sprintf(szBuf, "%d", nValue);
    m_papszParams = CSLSetNameValue(m_papszParams, pszName, szBuf);
}

/**********************************************************************
 *                       OWTGDChartURL::SetURLParam()
 *
 * Another version of the same function that takes a stringlist as input
 * and adds it to the list of params as a single string with the value
 * separated by semicolons (";").
 ********************************************************************** */
void OWTGDChartURL::SetURLParam(const char *pszName, char **papszValues,
                                char cDelim /* = ';' */)
{
    char *pszBuf, szDelim[2] = ";";
    int  i;
    unsigned int nTotLen = 0;

    if (papszValues == NULL)
        return;

    szDelim[0] = cDelim;

    /*---------------------------------------------------------------- 
     * Compute buffer size required to hold all values.
     * This is the total of 
     *   (length of each values + 1 char delimiter + 2 optional quotes)
     *----------------------------------------------------------------*/
    for(i=0; papszValues[i] != NULL; i++)
        nTotLen += strlen(papszValues[i]) + 1 + 2;
    nTotLen--;

    pszBuf = (char*)CPLMalloc(nTotLen+1*sizeof(char));
    pszBuf[0] = '\0';

    for(i=0; papszValues[i] != NULL; i++)
    {
        if (i != 0)
            strcat(pszBuf, szDelim);
        /*------------------------------------------------------------ 
         * If the value contains the delimiter character, then enclose
         * it inside double quotes.
         * __TODO__ This implies that a value cannot contain the double
         *          quote char and we'll have to find a way to escape 
         *          them at some point.
         *------------------------------------------------------------*/
        if (strstr(papszValues[i], szDelim))
        {
            strcat(pszBuf, "\"");
            strcat(pszBuf, papszValues[i]);
            strcat(pszBuf, "\"");
        }
        else
        {
            strcat(pszBuf, papszValues[i]);
        }
    }

    CPLAssert(strlen(pszBuf) <= nTotLen);

    m_papszParams = CSLSetNameValue(m_papszParams, pszName, pszBuf);

    CPLFree(pszBuf);
}

/**********************************************************************
 *                       OWTGDChartURL::GetParams()
 *
 * Get parameters.
 ********************************************************************** */
char **OWTGDChartURL::GetParams()
{
    return m_papszParams;
}
