/**********************************************************************
 * $Id: owtlog.c,v 1.2 2002/11/13 19:32:35 julien Exp $
 *
 * Name:     owtlog.c
 * Project:  OWT base library
 * Purpose:  Take care of writing to a log file
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1998, 1999, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: owtlog.c,v $
 * Revision 1.2  2002/11/13 19:32:35  julien
 * include <time.h> needed on RH 8.0 gcc 3.2
 *
 * Revision 1.1  1999/05/28 17:05:46  daniel
 * Initial revision
 *
 **********************************************************************/

#include "owtmisc.h"

#include <sys/timeb.h>

#ifdef _WIN32
#  include <time.h>
#else
#  include <time.h>
#  include <sys/time.h>
#endif

/*=====================================================================
                OWT functions to Log messages to a file
 =====================================================================*/

static FILE *gfpLog = NULL;
static int   gnLogLevel = OWTLOG_ALL;   /* By default: log everything */

/**********************************************************************
 *                       OWTLogStart()
 *
 * Start sending OWTLog() messages to specified output file.
 *
 * Returns 0 on success, or -1 on error.
 **********************************************************************/
int     OWTLogStart(const char *pszFname)
{
    OWTLogStop();

    gfpLog = fopen(pszFname, "a");

    return (gfpLog == NULL);
}

/**********************************************************************
 *                       OWTLogStop()
 *
 * Stop writing OWTLog() messages and close current log output file.
 **********************************************************************/
void    OWTLogStop()
{
    if (gfpLog)
    {
        fclose(gfpLog);
        gfpLog = NULL;
    }
}

/**********************************************************************
 *                       OWTLogSetLevel()
 *
 * Set new log level, returns previous log level value.
 **********************************************************************/
int     OWTLogSetLevel(int nNewLevel)
{
    int nPrevLevel;

    nPrevLevel = gnLogLevel;
    gnLogLevel = nNewLevel;

    return nPrevLevel;
}

/**********************************************************************
 *                       OWTLog()
 *
 * Send message to log file if log file has been previously opened with 
 * OWTLogStart() and specified log level is <= current log level.
 **********************************************************************/
int     OWTLog(int nLevel, const char *pszFmt, ...)
{
    int         nStatus = 0;

    if (gfpLog && nLevel <= gnLogLevel)
    {
        /*-------------------------------------------------------------
         * Output log message, preceded by current time
         *------------------------------------------------------------*/
        va_list       args;

#ifndef _WIN32
        /* __TODO__ We need a new implementation for WIN32 */
        struct tm     *psTM;
        struct timeval  tv;
        struct timezone tz;
        char          szBuf[100] = "";

        gettimeofday(&tv, &tz);
        psTM = localtime(&(tv.tv_sec));  /* Extract date/hour/min components */
        strftime(szBuf, 99, "%b %d %H:%M:%S", psTM);
        fprintf(gfpLog, "%s.%ld - ", szBuf,  tv.tv_usec/1000);
#endif

        va_start(args, pszFmt);
        nStatus = vfprintf(gfpLog, pszFmt, args);
        va_end(args);

        fprintf(gfpLog, "\n");

        fflush(gfpLog);
    }

    return nStatus;
}






