/**********************************************************************
 * $Id: owttags.c,v 1.3 1999/05/28 16:51:18 daniel Exp $
 *
 * Name:     owttags.c
 * Project:  OWT HTML manipulation library
 * Purpose:  OWT Tags manipulation functions.
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1998, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: owttags.c,v $
 * Revision 1.3  1999/05/28 16:51:18  daniel
 * Use EQUALN instead of strnicmp.
 *
 * Revision 1.2  1998/12/09 02:08:03  daniel
 * Converted to use portability library
 *
 * Revision 1.1  1998/12/04 19:15:20  daniel
 * Initial revision
 *
 * Revision 1.1  1998/11/05 04:36:47  daniel
 * Initial revision
 *
 **********************************************************************/

#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "owttags.h"

/*=====================================================================
                    OWT Tags manipulation functions.
 =====================================================================*/

/**********************************************************************
 *                       OWTPreprocessTags()
 *
 * Scan a StringList containing HTML and prepare it for use with the
 * other OWTTags functions.
 *
 * The main thing is that OWTTags have to be on separate lines to 
 * facilitate further manipulations.
 *
 * Returns the updated HTML Doc.
 **********************************************************************/
char **OWTPreprocessTags(char **papszDoc)
{
    int     iLine, nLenTagEnd;
    char    *pszTag;

    nLenTagEnd = strlen(OWTTAG_END);

    for(iLine=0; papszDoc && papszDoc[iLine] != NULL; iLine++)
    {
        pszTag = strstr(papszDoc[iLine], OWTTAG_START);
        if (pszTag && pszTag != papszDoc[iLine])
        {
            /* Split current line at tag location and insert the rest 
             * of the line in a new line after the current one.
             */
            papszDoc = CSLInsertString(papszDoc, iLine+1, pszTag);
            *pszTag = '\0';

            iLine++;
        }
        
        /* If the line contains a OWT Tag, then make sure that the tag ends
         * on this line as well.
         */
        if (pszTag)
        {
            pszTag = strstr(papszDoc[iLine], OWTTAG_END);
            if (pszTag == NULL)
            {
                /* ERROR: the OWT tag is not all contained in the current line
                 */
                CPLError(CE_Warning, CPLE_AppDefined, 
                        "OWTPreprocessTags(): Incomplete OWT Tag in line %d. "
                        "Note that OWT Tags cannot be split on several lines.",
                        iLine+1);
                return papszDoc;
            }
            else if (pszTag[nLenTagEnd] != '\0')
            {
                /* Move anything that follows the end of the tag to a
                 * new line
                 */
                papszDoc = CSLInsertString(papszDoc, iLine+1, 
                                              pszTag + nLenTagEnd);
                pszTag[nLenTagEnd] = '\0';
            }

        }/* if pszTag */
    }/* for iLine*/

    return papszDoc;
}


/**********************************************************************
 *                       OWTGetNextTagInfo()
 *
 * Find the location of the next OWT Tag in the document, starting
 * at nFirstLine, and going no further than nLastLine (inclusive).
 *
 * Fills and returns a new OWTTagInfo structure containing all info about
 * the tag that was found, or NULL if no OWT Tag was found or if errors
 * happened.
 *
 * OWTFreeTagInfo() will eventually have to be called to free the 
 * returned OWTTagInfo structure.
 **********************************************************************/
OWTTagInfo  *OWTGetNextTagInfo(char **papszDoc, int nFirstLine, int nLastLine)
{
    OWTTagInfo  *psTagInfo = NULL;
    int         iLine, nLen, nTotalLines, nLenTagStart;
    char        *pszLine, *pszTok;

    /* If nFirstLine and/or nLastLine have negative values then we 
     * ignore them.
     */
    nTotalLines = CSLCount(papszDoc);
    if (nFirstLine < 0)
        nFirstLine = 0;
    if (nLastLine < 0 || nLastLine >= nTotalLines)
        nLastLine = nTotalLines -1;

    nLenTagStart = strlen(OWTTAG_START);

    for(iLine=nFirstLine; iLine <= nLastLine; iLine++)
    {
        if (strncmp(papszDoc[iLine], OWTTAG_START, nLenTagStart) == 0
            && papszDoc[iLine][nLenTagStart] != '/')
        {
            /* We found one... alloc a new OWTTagInfo and
             * parse the name/type/options parts of the tag
             */
            psTagInfo = (OWTTagInfo*)calloc(1, sizeof(OWTTagInfo));
            if (psTagInfo == NULL ||
                (pszLine = strdup(papszDoc[iLine]+nLenTagStart)) == NULL)
            {
                /* Error: Out of memory!!! */
                CPLError(CE_Fatal, CPLE_OutOfMemory, 
                        "OWTGetNextTagInfo(): Unable to allocate memory.");
                return NULL;
            }
            /* Mark the end of the tag with a special character ('\n') to 
             * facilitate parsing.
             */
            nLen = strlen(pszLine) - strlen(OWTTAG_END);
            pszLine[nLen] = '\n';
            pszLine[nLen+1] = '\0';
            /* Extract the tag name
             */
            pszTok = strtok(pszLine, " \t\n");
            if (pszTok)
            {
                psTagInfo->pszName = strdup(pszTok);
                pszTok = strtok(NULL, " \t\n");
            }
            /* Extract tag type
             */
            if (pszTok)
            {
                psTagInfo->pszType = strdup(pszTok);
                pszTok = strtok(NULL, "\n");
            }
            /* And finally extract the options part
             */
            if (pszTok)
            {
                psTagInfo->pszOptions = strdup(pszTok);
            }
            /* Check for errors
             */
            if (psTagInfo->pszName == NULL || psTagInfo->pszType == NULL)
            {
                /* Error: Malformed tag, or less likely: Out of memory */
                CPLError(CE_Failure, CPLE_AppDefined, 
                        "OWTGetNextTagInfo(): Failed to parse tag in line %d. "
                        "Make sure tag name and type are properly specified.",
                        iLine + 1);
                OWTFreeTagInfo(psTagInfo);
                free(pszLine);
                return NULL;
            }

            /* Set the start line number in the psTagInfo
             */
            psTagInfo->nFirstLine = iLine;

            /* Find the end tag for this block
             */
            sprintf(pszLine, "%s/%s", OWTTAG_START, psTagInfo->pszName);
            nLenTagStart = strlen(pszLine);
            iLine++;
            while(papszDoc[iLine] != NULL &&
                  ! EQUALN(papszDoc[iLine], pszLine, nLenTagStart))
                  iLine++;
            if (papszDoc[iLine] == NULL)
            {
                CPLError(CE_Failure, CPLE_AppDefined, 
                        "OWTGetNextTagInfo(): Expected tag \"%s%s\" not found.",
                        pszLine, OWTTAG_END);
                OWTFreeTagInfo(psTagInfo);
                free(pszLine);
                return NULL;
            }
            psTagInfo->nNumLines = iLine - psTagInfo->nFirstLine +1;

            free(pszLine);
        }
    }

    return psTagInfo;
}


/**********************************************************************
 *                       OWTFreeTagInfo()
 *
 * Release all memory used by a OWTTagInfo, including the structure
 * itself.
 **********************************************************************/
void OWTFreeTagInfo(OWTTagInfo *psTagInfo)
{
    if (psTagInfo)
    {
        if (psTagInfo->pszName)
            free(psTagInfo->pszName);
        if (psTagInfo->pszType)
            free(psTagInfo->pszType);
        if (psTagInfo->pszOptions)
            free(psTagInfo->pszOptions);
        free(psTagInfo);
    }
}

/**********************************************************************
 *                       OWTPrintTagInfo()
 *
 * Print the contents of a OWTTagInfo struct to stdout... useful for
 * debugging.
 **********************************************************************/
void OWTPrintTagInfo(OWTTagInfo *psTagInfo)
{
    if (psTagInfo == NULL)
    {
        printf("(null OWTTagInfo)\n");
    }
    else
    {
        printf("pszName    = [%s]\n", 
                 psTagInfo->pszName?psTagInfo->pszName:"(null)");
        printf("pszType    = [%s]\n", 
                 psTagInfo->pszType?psTagInfo->pszType:"(null)");
        printf("pszOptions = [%s]\n", 
                 psTagInfo->pszOptions?psTagInfo->pszOptions:"(null)");
        printf("nFirstLine = %d\n", psTagInfo->nFirstLine);
        printf("nNumLines  = %d (including start/end tag lines)\n", 
                 psTagInfo->nNumLines);
    }
}
